from collections import defaultdict
from functools import wraps
from threading import Thread, current_thread

from typing_extensions import overload, Callable, Any, Union, Dict, List, Tuple, Optional, TypeVar, ParamSpec, TypeAlias

from .pyprof import Profiler

T = TypeVar('T')
P = ParamSpec('P')
FuncType: TypeAlias = Callable[P, T]


class ProfilerProxy:
    active_proxy: Dict[Thread, List[Tuple["ProfilerProxy", "Profiler"]]] = defaultdict(list)

    def __init__(
            self, name: str, report_printer: Optional[Callable[[str], Any]] = None, flush=False,
            min_total_percent: float = 0., min_parent_percent: float = 0.
    ):
        self.name = name
        self.report_printer = report_printer
        self.flush = flush
        self.min_total_percent = min_total_percent
        self.min_parent_percent = min_parent_percent

    @classmethod
    def nearest_proxy(cls) -> Optional[Tuple['ProfilerProxy', Profiler]]:
        current_stack = cls.active_proxy[current_thread()]
        return current_stack[-1] if current_stack else None

    def __enter__(self) -> Profiler:
        profiler = Profiler(
            self.name,
            current_profiler(),
            flush=self.flush,
        )
        self.flush = False  # flush for the first tic-toc only
        self.active_proxy[current_thread()].append((self, profiler))
        profiler.__enter__()
        return profiler

    def __exit__(self, exc_type, exc_val, exc_tb):
        _, profiler = self.active_proxy[current_thread()].pop()
        profiler.__exit__(exc_type, exc_val, exc_tb)
        if self.report_printer is not None:
            self.report_printer(
                profiler.report_header() + profiler.report(
                    min_total_percent=self.min_total_percent,
                    min_parent_percent=self.min_parent_percent,
                )
            )

    def __call__(self, func: FuncType) -> FuncType:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper


@overload
def profile(
        arg: str, *, report_printer=None, flush: bool = False,
        min_total_percent: float = 0.,
        min_parent_percent: float = 0.,
) -> ProfilerProxy:
    ...


@overload
def profile(arg: FuncType) -> FuncType:
    ...


def profile(
        arg: Union[str, FuncType], *, report_printer=None,
        flush: bool = False,
        min_total_percent: float = 0.,
        min_parent_percent: float = 0.,
) -> Union[FuncType, ProfilerProxy]:
    # work as a context manager
    if isinstance(arg, str):
        return ProfilerProxy(
            arg, report_printer=report_printer, flush=flush,
            min_total_percent=min_total_percent,
            min_parent_percent=min_parent_percent,
        )

    func = arg

    # work as a decorator
    @wraps(func)
    def wrapper(*args, **kwargs):
        with ProfilerProxy(
                name=func.__qualname__,
                report_printer=report_printer,
                flush=flush,
        ):
            return func(*args, **kwargs)

    return wrapper


def current_profiler() -> Optional[Profiler]:
    """
    :return: return the current profiler in the profiler stack defined by profile()
    """
    p = ProfilerProxy.nearest_proxy()
    return p[1] if p else None


__all__ = ['profile', 'current_profiler']
